import traceback
from logging import Logger
from threading import Thread
from time import sleep
from typing import Callable


def daemon(function: Callable, interval: float, logger: Logger = None) -> Callable:
    def decorator(*args, **kwargs):
        while True:
            try:
                function(*args, **kwargs)
            except Exception as exception:
                if logger is None:
                    traceback.print_exc()
                else:
                    logger.exception(exception)
            sleep(interval)

    return decorator


def parallel(function: Callable, concurrency: int, logger: Logger = None) -> Callable:
    def decorator(*args, **kwargs):
        def worker():
            try:
                function(*args, **kwargs)
            except Exception as exception:
                if logger is not None:
                    logger.exception(exception)
                raise

        threads = [Thread(target=worker) for _ in range(concurrency)]

        for thread in threads:
            thread.start()
        for thread in threads:
            thread.join()

    return decorator



from typing import Callable, List

from django.apps import apps
from django.utils.module_loading import import_string


def apps_lookup(template: str) -> List[str]:
    return [template % a.module.__name__ for a in apps.get_app_configs()]


def import_script(script_path: str) -> Callable:
    lookup_paths, module_path, script_name = [], None, script_path

    if '.' in script_name:
        module_path, script_name = script_path.rsplit('.', 1)

    if module_path is None:
        lookup_paths += apps_lookup(f'%s.scripts.{script_name}.run')
        lookup_paths += apps_lookup(f'%s.scripts.{script_name}')
    else:
        lookup_paths.append(f'{module_path}.{script_name}.run')
        lookup_paths.append(f'{module_path}.{script_name}')

        lookup_paths.append(f'{module_path}.scripts.{script_name}.run')
        lookup_paths.append(f'{module_path}.scripts.{script_name}')

        lookup_paths += apps_lookup(f'%s.{module_path}.scripts.{script_name}.run')
        lookup_paths += apps_lookup(f'%s.{module_path}.scripts.{script_name}')

    for lookup_path in lookup_paths:
        try:
            run_script = import_string(lookup_path)
            assert isinstance(run_script, Callable)
        except (ImportError, AssertionError):
            pass
        else:
            return run_script

    raise ImportError(f'Cannot find {script_path}')



import logging
from argparse import ArgumentParser

from django.core.management import BaseCommand

from django_runscript.decorators import daemon, parallel
from django_runscript.utilities import import_script


class Command(BaseCommand):
    def add_arguments(self, parser: ArgumentParser):
        parser.add_argument('script_path', nargs='?')
        parser.add_argument('-d', '--daemon', action='store_true')
        parser.add_argument('-i', '--interval', default=5, type=float)
        parser.add_argument('-c', '--concurrency', default=1, type=int)
        parser.add_argument('-l', '--logger', type=str)

    def handle(self, *args, script_path: str, **options):
        if script_path is None:
            raise ValueError('Script path required.')

        run_script = import_script(script_path)

        if options['logger'] is not None:
            options['logger'] = logging.getLogger(options['logger'])

        if options['daemon']:
            run_script = daemon(run_script, options['interval'], options['logger'])

        if options['concurrency'] > 1:
            run_script = parallel(run_script, options['concurrency'], options['logger'])

        run_script()



from distutils.core import setup

setup(
    name='django-runscript',
    packages=['django_runscript'],
    package_data={'django_runscript': ['management/commands/*.py']},
    version='0.1.1',
    license='MIT',
    description='Django command for running custom scripts with django environment.',
    author='retxxxirt',
    author_email='retxxirt@gmail.com',
    url='https://github.com/retxxxirt/django-runscript',
    keywords=['django runscript', 'django', 'django scripts', 'django management'],
    tests_require=(tests_require := ['django>=3.0,<3.1']),
    extras_require={'test': tests_require}
)
